import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from scipy.interpolate import LinearNDInterpolator
from datetime import datetime

class GravitySurvey():
    """
    Parameters
    ----------
        ranges: list[list] [[x_min, x_max], [y_min, y_max], [z_min, z_max]]
            Defines the range of the survey in the 3 different directions
            If only a single value is given for the z values, that dimension is assumed to be a constant value (no elevation changes) (default:None)
        survey_coordinates: np.array
            Expected shape is [num_points, 3] (default:None)
        noise_scale : float
            The standard deviation of gaussian noise (default: None)
        survey_shape: list
            The number of survey points on a grid, if a grid is used, or the number of survay points in total if a random arrangement is used. (default: None)
        noise_on_location_scale: float
            The standard deviation of the noise added to the survey locations
        noise: array
            Array survey noise. Can be made from noise_scale
        noise_on_location: array
            Array of noise to be added to the survey_coordinates. 
        gravity: array
            Array of gravity measurements.
    """
    def __init__(self, ranges=None, survey_coordinates=None, noise_scale=None, noise_on_location_scale=0.0, survey_shape=None):
        self.ranges = ranges
        self.survey_coordinates = survey_coordinates
        self.noise_scale = noise_scale
        self.noise_on_location_scale = noise_on_location_scale
        self.noise = None
        self.noise_on_location = None
        self.survey_shape = survey_shape
        self.gravity = None
        self.units = r'mGal'

    def __setattr__(self, name, value):
        if name == 'ranges':
            if value is not None:
                if not isinstance(value, list):
                    raise ValueError("ranges is not a list.")
                if len(value) != 3:
                    raise ValueError("The shape of ranges attribute is not as required.")
        if name == 'survey_coordinates':
            if value is not None:
                if not isinstance(value, np.ndarray):
                    raise ValueError("survey_coordinates has to be of type ndarray.")
                self.num_points = np.shape(value)[0]
        if name == 'gravity':
            if value is not None:
                if not isinstance(value, np.ndarray):
                    raise ValueError("Gravity has to be of type ndarray.")
                if self.survey_coordinates is not None:
                    if np.shape(value)[0] != np.shape(self.survey_coordinates)[0]:
                        raise ValueError("The shape of the gravity and the survey has to agree.")
        if name == 'noise':
            if value is not None:
                if not isinstance(value, np.ndarray):
                    raise ValueError("Noise has to be of type ndarray.")
        super().__setattr__(name, value)

    def make_survey(self, survey_shape=None):
        """
        Method to make the survey points.
        Parameters
        ----------
            survey_shape: list
                the type of survey configuration is inferred from the lenght of the list
                len = 1 --> randomised survey locations
                len = 2 --> 2D grid, [number of points in the x dimension, y dimension]
                len = 3 --> 3D grid, [x, y, z]
        Output
        ------
            survey_coordinates: np.ndarray
                Shape: [number of survey locations, (x,y,z) coordinates]
        """
        if self.ranges is None:
            raise ValueError('The ranges of the survey coordinates are not defined.')
        if survey_shape is not None:
            self.survey_shape = survey_shape
        x_min, x_max = self.ranges[0]
        y_min, y_max = self.ranges[1]

        self.make_noise_on_location()

        if self.survey_shape is None:
            raise ValueError('The survey_shape is not defined.')
        if len(self.survey_shape) == 2:
            z = self.ranges[2][0]
            dx = (x_max - x_min)/self.survey_shape[0]
            dy = (y_max - y_min)/self.survey_shape[1]
            x_cm = np.arange(x_min+dx/2, x_max+dx/2, dx)[:self.survey_shape[0]]
            y_cm = np.arange(y_min+dy/2, y_max+dy/2, dy)[:self.survey_shape[1]]

            X_cm, Y_cm, Z_cm = np.meshgrid(x_cm, y_cm, z, indexing='ij')
            X_cm, Y_cm, Z_cm = X_cm.ravel(), Y_cm.ravel(), Z_cm.ravel()
            survey_coordinates = np.c_[X_cm, Y_cm, Z_cm]
            survey_coordinates[:,:2] = survey_coordinates[:,:2] + self.noise_on_location[:,:2]
            self.survey_coordinates = survey_coordinates
        elif len(self.survey_shape) == 3:
            if len(self.ranges[2]) != 2:
                raise ValueError("range is not defined for the z dimension.")
            z_min, z_max = self.ranges[2]
            dx = (x_max - x_min)/self.survey_shape[0]
            dy = (y_max - y_min)/self.survey_shape[1]
            dz = (z_max - z_min)/self.survey_shape[2]
            x_cm = np.arange(x_min+dx/2, x_max+dx/2, dx)[:self.survey_shape[0]]
            y_cm = np.arange(y_min+dy/2, y_max+dy/2, dy)[:self.survey_shape[1]]
            z_cm = np.arange(z_min+dy/2, z_max+dy/2, dz)[:self.survey_shape[2]]

            X_cm, Y_cm, Z_cm = np.meshgrid(x_cm, y_cm, z_cm, indexing='ij')
            X_cm, Y_cm, Z_cm = X_cm.ravel(), Y_cm.ravel(), Z_cm.ravel()
            survey_coordinates = np.c_[X_cm, Y_cm, Z_cm]
            survey_coordinates = survey_coordinates + self.noise_on_location
            self.survey_coordinates = survey_coordinates
        elif len(self.survey_shape) == 1:
            x = np.random.uniform(self.ranges[0][0], self.ranges[0][1], size=self.survey_shape)
            y = np.random.uniform(self.ranges[1][0], self.ranges[1][1], size=self.survey_shape)
            if len(self.ranges[2]) == 2:
                z = np.random.uniform(self.ranges[2][0], self.ranges[2][1], size=self.survey_shape)
            else:
                z = np.ones(self.survey_shape)*self.ranges[2][0]
            survey_coordinates = np.vstack([x, y, z]).T
            survey_coordinates[:,:2] = survey_coordinates[:,:2] + self.noise_on_location[:,:2]
            self.survey_coordinates = survey_coordinates
        else:
            raise ValueError('The given survey_shape cannot be interpreted')
        return survey_coordinates

    def make_noise(self, noise_scale=None):
        """
        Generates an array of gaussian noise based on the given noise_scale.
        Parameters
        ----------
            noise_scale: float
                The standard deviation of the gaussian noise. If None, then taken from the class attributes.
        Output
        ------
            noise: np.ndarray
                Same number of elements ad survey points in the class. Contains simulated noise with the desired standard deviation.
        """
        if noise_scale is not None:
            self.noise_scale = noise_scale
        else:
            if self.noise_scale is None:
                raise ValueError("noise scale was not added as an input or attribute to the class.")
        num_points = self.get_number_of_surveypoints()
        self.noise = np.random.normal(loc=0.0, scale=self.noise_scale, size=num_points)
        return self.noise

    def make_noise_on_location(self, noise_on_location_scale=None):
        """
        Generates noise for the location of the survey points.
        Parameters
        ----------
            noise_on_location_scale: float
                If none, taken from the class. Otherwise defines the standard deviation of the noise added to the survey locations.
        Output
        ------
            noise_on_location: np.ndarray
                Returns an array of the shape [num survey points, 3] with simulated noise. Can be added to the survey_coordinates.
        """
        if noise_on_location_scale is not None:
            self.noise_on_location_scale = noise_on_location_scale
        else:
            if self.noise_on_location_scale is None:
                raise ValueError("noise scale was not added as an input or attribute to the class.")
        num_points = self.get_number_of_surveypoints()
        self.noise_on_location = np.random.normal(loc=0.0, scale=self.noise_on_location_scale, size=(num_points, 3))
        return self.noise_on_location

    def get_number_of_surveypoints(self):
        """
        Calculates the number of survey points from the survey_shape or survey_coordinates.
        Output
        -----
            num_points: int
                the number of survey points
        """
        if self.survey_coordinates is not None:
            num_points = np.shape(self.survey_coordinates)[0]
        elif self.survey_shape is not None:
            num_points = 1
            for i in self.survey_shape:
                num_points = num_points*i
        else:
            raise ValueError("Not enough information is given. Provide either survey_shape or survey_coordinates to the class")
        return num_points

    def add_noise(self):
        """
        Adds simulated noise to the gravity measurements.
        Output
        ------
            np.ndarray
                gravity+noise
        """
        if self.noise is None:
            raise ValueError("Noise has not been defined.")
        return self.gravity + self.noise

    def snr(self, noise_scale=None):
        """
        Compute the SNR. NOT USED!
        Parameters
        ----------
            noise_scale: float
                The standard deviation of the survey noise. If None, then taken from the class attributes.
        Output
        ------
            snr: float
                std of the survey data/ std of the noise
        """
        if noise_scale is None:
            noise_scale = self.noise_scale
            if noise_scale is None:
                raise ValueError("Set the noise_scale class attribute or function input first")
        num_points = self.get_number_of_surveypoints()
        snr = np.sqrt(np.sum((self.gravity-np.mean(self.gravity))**2)/num_points)/(noise_scale)
        return snr

    def get_spectrum(self):
        g_f = np.fft.fft2(np.reshape(self.gravity, self.survey_shape))
        asd = np.sqrt(np.abs(g_f)**2)
        return asd

    def interpolate_data_onto_survey_coordinates(self, data=None, original_grid=None, func=None, remove_min=True):
        if func is None:
            if data is None:
                raise ValueError('Need to either provide the fitted interpolation or the data and the grid to fit to.')
            if original_grid is None:
                raise ValueError('Need to either provide the fitted interpolation or the data and the grid to fit to.')
            coords = np.reshape(original_grid, (np.shape(original_grid)[0]*np.shape(original_grid)[1], np.shape(original_grid)[2]))
            start_time = datetime.now()
            func = LinearNDInterpolator(coords, data)
        g_new = func(self.survey_coordinates[:,0], self.survey_coordinates[:,1])
        if remove_min:
            self.gravity = g_new-np.min(g_new)
        else:
            self.gravity = g_new
        return self.gravity

    #Plotting
    def plot_pixels(self, filename='survey.png', include_noise=False):
        """
        Creates a simple pixelised image of the survey. Can only be done for gridded data.
        Parameters
        ----------
            filename: str
                The location where the image is saved.
            include_noise: bool
                If True, the noise is added to the survey before plotting.
        """
        if len(self.survey_shape) == 1:
            raise ValueError("This function is only available for gridded data points.")
        if self.gravity is None:
            raise ValueError("Compute the gravity first.")
        if include_noise is True:
            if self.noise is None:
                self.make_noise()
            plot_data = self.add_noise()
        else:
            plot_data = self.gravity
        if self.survey_coordinates is not None:
            extent = (np.min(self.survey_coordinates[:,0]), np.max(self.survey_coordinates[:,0]), np.max(self.survey_coordinates[:,1]), np.min(self.survey_coordinates[:,1]))
            plt.imshow(np.rot90(np.reshape(plot_data, self.survey_shape)), extent=extent)
            plt.xlabel('x')
            plt.ylabel('y')
        elif self.ranges is not None:
            extent = (self.ranges[0][0], self.ranges[0][1], self.ranges[1][1], self.ranges[1][0])
            plt.imshow(np.rot90(np.reshape(plot_data, self.survey_shape)), extent=extent)
            plt.xlabel('x')
            plt.ylabel('y')
        else:
            print('Need ranges or survey_coordinates to compute extent of survey area. Plotting without.')
            plt.imshow(np.rot90(np.reshape(plot_data, self.survey_shape)))
        # the extent and order of coordinates has been validated
        plt.colorbar(label=self.units)
        plt.savefig(filename)
        plt.close()

    def plot_contours(self, filename='survey.png', include_noise=False, plot_survey_coordinates=True):
        """
        Creates a contour plot of the survey.
        Parameters
        ----------
            filename: str
                The location where the image is saved.
            include_noise: bool
                If True, the noise is added to the survey before plotting.
        """
        if self.gravity is None:
            raise ValueError("Compute the gravity first.")
        if include_noise is True:
            if self.noise is None:
                self.make_noise()
            plot_data = self.add_noise()
        else:
            plot_data = self.gravity
        levels = np.linspace(self.gravity.min(), self.gravity.max(), 20)
        cmap = 'plasma'
        norm = matplotlib.colors.Normalize(vmin=self.gravity.min(), vmax=self.gravity.max())
        fig, ax = plt.subplots()
        if plot_survey_coordinates:
            ax.plot(self.survey_coordinates[:,0], self.survey_coordinates[:,1], 'o', markersize=2, color='black')
        ax.tricontourf(self.survey_coordinates[:,0], self.survey_coordinates[:,1], plot_data, levels=levels, cmap=cmap, norm=norm)
        ax.set(xlim=(np.min(self.survey_coordinates[:,0]), np.max(self.survey_coordinates[:,0])), ylim=(np.min(self.survey_coordinates[:,1]), np.max(self.survey_coordinates[:,1])))
        ax.set(ylabel='y')
        ax.set(xlabel='x')
        ax.set_aspect(aspect='equal')
        plt.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap), ax=ax, label=self.units)
        plt.savefig(filename)
        plt.close()

    def change_reference_frame(self, coordinate_centres=[0,0]):
        """
        Chenges the frame of reference of the coordiante system.
        Parameters
        ---------
            coordinate_centres: list of floats
                This should be in the current frame,
                Defines the centre point of the new coordinate system
        """
        x, y = coordinate_centres
        self.survey_coordinates[:,0] = self.survey_coordinates[:,0] - x
        self.survey_coordinates[:,1] = self.survey_coordinates[:,1] - y
        self.ranges[0][0], self.ranges[0][1] = self.ranges[0][0]-x, self.ranges[0][1]-x
        self.ranges[1][0], self.ranges[1][1] = self.ranges[1][0]-y, self.ranges[1][1]-y

    def scale_by_width(self, width=1.0):
        if self.gravity is not None:
            self.gravity = self.gravity/width

        if self.noise_scale is not None:
            self.noise_scale = self.noise_scale/width
        if self.noise is not None:
            self.noise = self.noise/width

        if self.survey_coordinates is not None:
            self.survey_coordinates = self.survey_coordinates/width # scales z too!
        if self.ranges is not None:
            new_ranges = []
            for r in self.ranges:
                new_range_element = [rr/width for rr in r]
                new_ranges.append(new_range_element)
            self.ranges = new_ranges

        if self.noise_on_location_scale is not None:
            self.noise_on_location_scale = self.noise_on_location_scale/width
        if self.noise_on_location is not None:
            self.noise_on_location = self.noise_on_location/width

class GradientSurvey(GravitySurvey):
    def __init__(self, ranges=None, survey_coordinates=None, noise_scale=None, noise_on_location_scale=0.0, survey_shape=None):
        GravitySurvey.__init__(self, ranges, survey_coordinates, noise_scale, noise_on_location_scale, survey_shape)
    #Plotting
        self.units = r'E'

